from __future__ import annotations

from itertools import product as cartprod
from copy import deepcopy

class Piece:
    ''' Basic functionality of TicTacToe pieces. '''
    pieces = {
        0: '-',    # 0 is hardcoded to be empty piece
        1: 'X',
        2: 'O',
        # 3: 'Y' 
    }
    playernumber = len(pieces)-1
    
    def __init__(self, code: int = 0) -> None:
        ''' Initialize piece. '''
        if not isinstance(code, int):
            raise ValueError('Piece cant be initialized with non-int.')
        self.name = Piece.pieces[code]
        self.code = code

    @staticmethod
    def reverse(piece: Piece) -> Piece:
        ''' Return next piece in Piece.pieces. '''
        if not isinstance(piece, Piece):
            raise ValueError('Cant reverse non-piece object')
        if piece.code == Piece.playernumber:
            return Piece(1)
        else:
            return Piece(piece.code+1)

    @staticmethod
    def previous(piece: Piece) -> Piece:
        ''' Return previous piece in Piece.pieces.'''
        if not isinstance(piece, Piece):
            raise ValueError('Cant reverse non-piece object')
        if piece.code == 1:
            return Piece(Piece.playernumber)
        else:
            return Piece(piece.code-1)

    def __eq__(self, other: Piece) -> bool:
        if not isinstance(other, Piece):
            return NotImplemented
        return self.code == other.code

    def __int__(self) -> int:
        return self.code

    def __str__(self) -> str:
        if hasattr(self.name, '__str__'):
            return self.name
        raise TypeError('This piece cannot be converted to str')

class Board:
    '''
    Only basic functionality of Board. There is no use of Tic Tac Toe game
    logic like win/draw, game ending, turn etc. Game logic is implemented in
    child class TicTacToe.
    '''

    def __init__(self, m=3, n=3, infinite=False) -> None:
        '''
        Initialize a board.
        (m, n) are dimensions of the board: n is number of rows, m is number of columns.
        infinite makes board an infinite matrix for with non-negative indexes (not actually infinite, but with dynamically changing dimensions).
        '''
        self.infinite = infinite
        self.m = m
        self.n = n
        self.board = [[Piece()]*m for _ in range(n)]

    def increase_dimensions(self, to_dim) -> None:
        ''' Scale up dimensions to (s, s). '''

        newboard = Board(to_dim, to_dim, infinite=True)

        for (x, y) in newboard.moves_coords():
            if (x, y) in self.moves_coords():
                newboard[x, y] = self[x, y]
            else:
                newboard[x, y] = Piece()

        self.n, self.m = to_dim, to_dim
        self.board = newboard.board.copy()

    def columns(self) -> list:
        ''' Return columns for checking win. '''
        board = self.board
        return [list(field) for field in zip(*board)]

    def diagonals(self) -> list:
        ''' Return diagonals for checking win. '''
        diags = []
        for sm in range(self.m+self.n-1):
            strip = []
            for (i, j) in self.moves_coords():
                if i+j == sm:
                    strip.append(self[i, j])
            if not len(strip):
                diags.append(strip)
        return diags

    def antidiagonals(self) -> list:
        ''' Return diagonals for checking win. '''
        # Rotating board by 90 degree and then getting its diagonals
        rotated_board = Board(m=self.m, n=self.n)
        rotated_board.board = list(zip(*self.board))[::-1]
        return rotated_board.diagonals()

    def rows(self):
        ''' Return rows for checking win. '''
        return self.board

    def lines(self) -> list:
        ''' Return all possible lines to be winned. '''
        return self.rows() + self.columns() + self.diagonals() + self.antidiagonals()

    def flat(self) -> list:
        ''' Return flat format of board. '''
        board = self.board
        return sum(board, [])
    
    def moves_coords(self) -> list:
        ''' Return all the fields on board'''
        return [(i,j) for (i,j) in cartprod(range(self.n), range(self.m))]

    def __getitem__(self, key):
        if not key in self.moves_coords() and not self.infinite:
            raise IndexError('Board index out of range.')

        if not key in self.moves_coords() and self.infinite:
            to_dim = max(key[0]+1, key[1]+1)
            self.increase_dimensions(to_dim)

        return self.board[key[0]][key[1]]

    def __setitem__(self, key, value):

        if not isinstance(key[0], int) or not isinstance(key[1], int):
            raise TypeError('Board index must be int.')

        if not key in self.moves_coords() and not self.infinite:
            raise IndexError('Board index out of range.')

        if not key in self.moves_coords() and self.infinite:
            to_dim = max(key[0]+1, key[1]+1)
            self.increase_dimensions(to_dim)

        if not isinstance(value, Piece):
            raise ValueError('Cannot set Board element to non-piece object.')
        
        self.board[key[0]][key[1]] = value

    def __str__(self) -> str:
        printed = ['\n']
        for row in self.board:
            for field in row:
                printed.append(str(field))
            printed.append('\n')
        return ''.join(printed)
